package pathfinder;

import java.util.ArrayList;
import java.util.Collections;
import pathfinder.Node;

public class AStar implements PathFinderInterface {
	
	private ArrayList closed = new ArrayList();
	private SortedList open = new SortedList();
	private MapInterface map;
	private int maxSearchDistance;
	private Node[][] nodes;
	private boolean allowDiagMovement;
	private HeuristicInterface heuristic;
	
	public int isLegal = 0;
	
	/**
	 * Two constructors: the first creates a pathfinder with a default heuristic if none is chosen
	 * The second is used when a specific heuristic is selected
	 */
	
	public AStar(MapInterface map, int maxSearchDistance, boolean allowDiagMovement) {
		this(map, maxSearchDistance, allowDiagMovement, new ClosestHeuristic());
	}
	
	public AStar(MapInterface map, int maxSearchDistance, boolean allowDiagMovement, HeuristicInterface heuristic) {
		this.heuristic = heuristic;
		this.map = map;
		this.maxSearchDistance = maxSearchDistance;
		this.allowDiagMovement = allowDiagMovement;
		
		nodes = new Node[map.getWidth()][map.getHeight()];
		for (int x=0; x < map.getWidth(); x++) {
			for (int y=0; y < map.getHeight(); y++) {
				nodes[x][y] = new Node(x,y);
			}
		}
	}	
	
	public Path findPath(Mover mover, int sx, int sy, int tx, int ty){
		
		/**
		 * Check if the initial move is blocked
		 */
		
		if (map.pathBlocked(mover, tx, ty)) {
			return null;
		}
		
		nodes[sx][sy].cost = 0;
		nodes[sx][sy].depth = 0;
		
		/**
		 * Add the starting location to the open list and empty the 
		 * closed list
		 */
		
		closed.clear();
		open.clear();
		open.add(nodes[sx][sy]);
		nodes[tx][ty].parent = null; 

		/**
		 * While there are still moves in the open list and we are within the maximum depth
		 */
		
		int maxDepth = 0;
		while ((maxDepth < maxSearchDistance) && (open.size() !=0)) {
			Node current = getFirstInOpen();
			if (current == nodes[tx][ty]) {
				//if current is the target node, we're done
				break;
			}
	
			/**
			 * if we havent found our target, remove from the open list and add to the closed list
			 */
			removeFromOpen(current);
			addToClosed(current);
			
			/**
			 * //consider each neighbour
			 */
			
			for (int x=-1; x<2; x++) {
				for (int y=-1; y<2; y++) {
					
					if ((x == 0) && (y == 0)) {
						continue;
						//Still on the same tile if we're here
					}
					
					if (!allowDiagMovement) {
						if ((x != 0) && (y !=0)) {
							continue;
							//If both x/y change we've made a diagonal move
						}
					}
					
					int xp = x + current.x;
					int yp = y + current.y;

					/**
					 * 	calculate the cost of reaching the neighbour	
					 */
					
					if (isValidLocation(mover, sx, sy, xp, yp)) {						
						float nextStepCost = current.cost + getMovementCost(mover, current.x, current.y, xp, yp);
						Node neighbour = nodes[xp][yp];
						map.pathNodeVisited(xp, yp);
						
						/**
						 * If the cost is less than the cost known for this location then remove it from the open or closed lists
						 *  (since we've now found a better route)						
						 */
						
						if (nextStepCost < neighbour.cost) {
							if (inOpenList(neighbour)) {
								removeFromOpen(neighbour);
							}
							
							if (inClosedList(neighbour)) {
								removeFromClosed(neighbour);
							}
						}
						
						/**
						 * If the location isn't in either the open or closed list then record the costs for the location and add it to the open list 
						 * (this means it'll be considered in the next search). Record how we got to this location				
						 */
						
						if (!inOpenList(neighbour) && !(inClosedList(neighbour))) {
							neighbour.cost = nextStepCost;
							neighbour.heuristic = getHeuristicCost(mover,xp,yp,tx,ty);
							maxDepth = Math.max(maxDepth, neighbour.setParent(current));
							addToOpen(neighbour);							
						}
					}
				}	
			}
		}
		
		/**
		 * If we're here and the target node does not have a parent, it means the search did not reach the target
		 * This means the search failed, so return null
		 */
		
		if (nodes[tx][ty].parent == null) {
			return null;
		}
		
		/**
		 * Otherwise, we've found a legal path so we can use the parent references to find the path back 
		 */
			
		Path path = new Path();
		Node target = nodes[tx][ty];
		while (target != nodes[sx][sy]) {
			path.prependStep(target.x, target.y);
			target = target.parent;
		}
				
		path.prependStep(sx, sy);		
		isLegal = 1;
		return path;
	}
	
	protected Node getFirstInOpen() {
		return (Node) open.first();
	}
	
	protected void addToOpen(Node node) {
		open.add(node);
	}
	
	protected boolean inOpenList(Node node) {
		return open.contains(node);
	}
	
	protected void removeFromOpen(Node node) {
		open.remove(node);	
	}
	
	protected void addToClosed(Node node) {
		closed.add(node);
	}
	
	protected boolean inClosedList(Node node) {
		return closed.contains(node);
	}
	
	protected void removeFromClosed(Node node) {
		closed.remove(node);
	}
	
	/**
	 * Check if a mover can be legally moved to a location
	 */
	protected boolean isValidLocation(Mover mover, int sx, int sy, int x, int y) {

		boolean invalid = (x < 0) || (y < 0) || (x >= map.getWidth()) || (y >= map.getHeight());
		
		if ((!invalid) && ((sx != x) || (sy != y))) {
			invalid = map.pathBlocked(mover, x, y);
		}
		
		return !invalid;
	}
	
	public float getMovementCost(Mover mover, int sx, int sy, int tx, int ty) {
		return map.getCost(mover, sx, sy, tx, ty);
	}

	public float getHeuristicCost(Mover mover, int x, int y, int tx, int ty) {
		return heuristic.getCost(map, mover, x, y, tx, ty);	
	}
	
	private class SortedList {
		
		private ArrayList list = new ArrayList();
		
		//Gets the first object in the list
		public Object first() {
			return list.get(0);
		}
		
		//Empties the list
		public void clear() {
			list.clear();
		}
		
		//Adds an object then sorts the list
		public void add(Object o) {
			list.add(o);
			Collections.sort(list);
		}
		
		//Removes an object
		public void remove(Object o) {
			list.remove(o);
		}
	
		//Gets the number of elements
		public int size() {
			return list.size();
		}
		
		//Checks if a value is in the list
		public boolean contains(Object o) {
			return list.contains(o);
		}
	}
}
